// File: blndfunc.cpp
// Auth: Brian Allen Vanderburg II
// Date: Tuesday, December 16, 2003
// Desc: Blend functions

/*
To add a blend function just create a function with the following form:

void __cdecl myBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	}

It is best to copy the normalBlendFunc and edit it where the comment says to

Then add the name of the blend and the function to the blendMethods list
at the bottom of this file:

{ "My Blend Function", myBlendFunc, "Extra 1 Desc", "Extra 2 Desc", myStartFunc },

myStartFunc has the following form:

int __cdecl myStartFunc(unsigned char *data, int ext1, int ext2, int w, int h)
	{
	if(data == NULL)
		return size_needed;

	....
	return 0;
	}

This function will get called twice.
When data is NULL, return the number of bytes the blend function
needs to allocate.  When data is not null, build the lookup table.
Return 0 for success or 1 for failure.
This can be NULL if a start func is not needed
See posterizeStartFunc for an example of it's usage


The last entry in the blendMethods should always be { NULL, NULL, NULL, NULL, NULL }

t_r, t_g, t_b - Represents the top pixel color
b_r, b_g, b_b - Represents the bottom pixel color
d_r, d_g, d_b - Represents the destination (blended) color
x, y - The current pixel position
w, h - The size of the image
frame - The current frame number
ext1, ext2 - Extra 1 and Extra two parameters
data - Data created by myStartFunc

You should blend the bottom and top pixels together and store the result
in the destination pixel.

*/


#include <stdlib.h>

#include "filter.h"
#include "blndfunc.h"


// This is a multiplication macro from GIMP
// It has been modified so the final value will still be asigned to t
#define INT_MULT(a, b, t)  ((t) = (a) * (b) + 0x80, (t) = ((((t) >> 8) + (t)) >> 8))



// Normal blend function
void __cdecl normalBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = t_r;
	*d_g = t_g;
	*d_b = t_b;
	}

// Darken blend function
void __cdecl darkenBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	// If bottom is darker, copy it to destination, otherwise copy the top
	if(b_r < t_r)
		*d_r = b_r;
	else
		*d_r = t_r;

	if(b_g < t_g)
		*d_g = b_g;
	else
		*d_g = t_g;

	if(b_b < t_b)
		*d_b = b_b;
	else
		*d_b = t_b;
	}

// Lighten blend function
void __cdecl lightenBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	// If bottom is lighter, copy it to destination, otherwise copy top
	if(b_r > t_r)
		*d_r = b_r;
	else
		*d_r = t_r;

	if(b_g > t_g)
		*d_g = b_g;
	else
		*d_g = t_g;

	if(b_b > t_b)
		*d_b = b_b;
	else
		*d_b = t_b;
	}

// Multiply blend function
void __cdecl multiplyBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	unsigned int tmp;

	*d_r = INT_MULT(t_r, b_r, tmp);
	*d_g = INT_MULT(t_g, b_g, tmp);
	*d_b = INT_MULT(t_b, b_b, tmp);
	}


// Screen blend function
void __cdecl screenBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	unsigned int tmp;

	*d_r = 255 - INT_MULT(255 - t_r, 255 - b_r, tmp);
	*d_g = 255 - INT_MULT(255 - t_g, 255 - b_g, tmp);
	*d_b = 255 - INT_MULT(255 - t_b, 255 - b_b, tmp);
	}

// Overlay blend function
void __cdecl overlayBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	unsigned int tmp;

	INT_MULT(2 * b_r, 255 - t_r, tmp);
	*d_r = INT_MULT(t_r, t_r + tmp, tmp);

	INT_MULT(2 * b_g, 255 - t_g, tmp);
	*d_g = INT_MULT(t_g, t_g + tmp, tmp);

	INT_MULT(2 * b_b, 255 - t_b, tmp);
	*d_b = INT_MULT(t_b, t_b + tmp, tmp);
	}

// Add blend function
void __cdecl addBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	unsigned int tmp;

	tmp = t_r + b_r;
	*d_r = (tmp > 255) ? 255 : tmp;

	tmp = t_g + b_g;
	*d_g = (tmp > 255) ? 255 : tmp;

	tmp = t_b + b_b;
	*d_b = (tmp > 255) ? 255 : tmp;
	}

// Subtract blend function
void __cdecl subtractBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	signed int tmp;

	tmp = t_r - b_r;
	*d_r = (tmp < 0) ? 0 : tmp;

	tmp = t_g - b_g;
	*d_g = (tmp < 0) ? 0 : tmp;

	tmp = t_b - b_b;
	*d_b = (tmp < 0) ? 0 : tmp;
	}

// Difference blend function
void __cdecl differenceBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	signed int tmp;

	tmp = t_r - b_r;
	*d_r = (tmp < 0) ? -tmp : tmp;

	tmp = t_g - b_g;
	*d_g = (tmp < 0) ? -tmp : tmp;

	tmp = t_b - b_b;
	*d_b = (tmp < 0) ? -tmp : tmp;
	}


// AND blend func
void __cdecl andBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = t_r & b_r;
	*d_g = t_g & b_g;
	*d_b = t_b & b_b;
	}

// OR blend func
void __cdecl orBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = t_r | b_r;
	*d_g = t_g | b_g;
	*d_b = t_b | b_b;
	}

// XOR blend func
void __cdecl xorBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = t_r ^ b_r;
	*d_g = t_g ^ b_g;
	*d_b = t_b ^ b_b;
	}

// Channel rotate left function
void __cdecl chanrolBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	int amount = ext1 & 0x07;

	if(amount < 0)
		amount = -amount;

	*d_r = (t_r << amount) + (t_r >> (8 - amount));
	*d_g = (t_g << amount) + (t_g >> (8 - amount));
	*d_b = (t_b << amount) + (t_b >> (8 - amount));
	}

// Pixel rotate left function
void __cdecl pixrolBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	int amount = ext1 % 24;
	unsigned long value;

	if(amount < 0)
		amount = -amount;

	value = (t_r << 16) + (t_g << 8) + t_b;

	value = (value << amount) + (value >> (24 - amount));

	*d_r = (unsigned char)((value & 0x00FF0000) >> 16);
	*d_g = (unsigned char)((value & 0x0000FF00) >> 8);
	*d_b = (unsigned char)(value & 0x000000FF);
	}

// Lines blend function
void __cdecl linesBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	if(ext1 < 1)
		ext1 = 1;

	if(ext2 < 1)
		ext2 = 1;

	if(x % (ext1 + 1) == 0 || y % (ext2 + 1) == 0)
		{
		*d_r = t_r;
		*d_g = t_g;
		*d_b = t_b;
		}
	else
		{
		*d_r = b_r;
		*d_g = b_g;
		*d_b = b_b;
		}
	}

// horizonal blend function
void __cdecl horzBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	if(ext1 < 1)
		ext1 = 1;

	if(y % (ext1 + 1) == 0)
		{
		*d_r = t_r;
		*d_g = t_g;
		*d_b = t_b;
		}
	else
		{
		*d_r = b_r;
		*d_g = b_g;
		*d_b = b_b;
		}
	}

// vertical blend function
void __cdecl vertBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	if(ext1 < 1)
		ext1 = 1;

	if(x % (ext1 + 1) == 0)
		{
		*d_r = t_r;
		*d_g = t_g;
		*d_b = t_b;
		}
	else
		{
		*d_r = b_r;
		*d_g = b_g;
		*d_b = b_b;
		}
	}

// rotate red blend function
void __cdecl rotredBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = t_r + (unsigned char)ext1;
	*d_g = t_g;
	*d_b = t_b;
	}

// rotate green blend function
void __cdecl rotgreenBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = t_r;
	*d_g = t_g + (unsigned char)ext1;
	*d_b = t_b;
	}

// rotate blue blend function
void __cdecl rotblueBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = t_r;
	*d_g = t_g;
	*d_b = t_b + (unsigned char)ext1;
	}

// posterize start function
int __cdecl posterizeStartFunc(unsigned char *data, int ext1, int ext2, int w, int h)
	{
	int pos;

	if(data == NULL)
		return 256; // We need 256 bytes

	if(ext1 <= 0)
		ext1 = 1;

	// build lookup table
	for(pos = 0; pos < 256; pos++)
		data[pos] = pos - (pos % ext1);

	return 0; // success
	}

// posterize blend function
void __cdecl posterizeBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	*d_r = data[t_r];
	*d_g = data[t_g];
	*d_b = data[t_b];
	}

// solarize blend function
void __cdecl solarizeBlendFunc(unsigned char t_r, unsigned char t_g, unsigned char t_b, unsigned char b_r, unsigned char b_g, unsigned char b_b, unsigned char *d_r, unsigned char *d_g, unsigned char *d_b, int x, int y, int w, int h, long frame, int ext1, int ext2, unsigned char *data)
	{
	if(ext1 < 0)
		ext1 = 0;

	if(t_r > ext1)
		*d_r = ~t_r;
	else
		*d_r = t_r;

	if(t_g > ext1)
		*d_g = ~t_g;
	else
		*d_g = t_g;

	if(t_b > ext1)
		*d_b = ~t_b;
	else
		*d_b = t_b;
	}



// Build the blending methods structure
// Keep this structure at the bottom of the file below all
// blend functions so you do not need to create prototypes
// for the blend functions

// Blending methods
BlendMethod blendMethods[] =
	{
		{ "Normal", normalBlendFunc, NULL, NULL, NULL },
		{ "Darken", darkenBlendFunc, NULL, NULL, NULL },
		{ "Lighten", lightenBlendFunc, NULL, NULL, NULL },
		{ "Multiply", multiplyBlendFunc, NULL, NULL, NULL },
		{ "Screen", screenBlendFunc, NULL, NULL, NULL },
		{ "Overlay", overlayBlendFunc, NULL, NULL, NULL },
		{ "Add", addBlendFunc, NULL, NULL, NULL },
		{ "Subtract", subtractBlendFunc, NULL, NULL, NULL },
		{ "Difference", differenceBlendFunc, NULL, NULL, NULL },
		{ "And", andBlendFunc, NULL, NULL, NULL },
		{ "Or", orBlendFunc, NULL, NULL, NULL },
		{ "Exclusive Or", xorBlendFunc, NULL, NULL, NULL },
		{ "Channel Rotate Left", chanrolBlendFunc, "Bits to rotate channel left (0 - 7)", NULL, NULL },
		{ "Pixel Rotate Left", pixrolBlendFunc, "Bits to rotate pixel left (0 - 23)", NULL, NULL },
		{ "Lines", linesBlendFunc, "Horizontal spacing", "Vertical spacing", NULL },
		{ "Horizonal Lines", horzBlendFunc, "Horizontal spacing", NULL, NULL },
		{ "Vertical Lines", vertBlendFunc, "Vertical spacing", NULL, NULL },
		{ "Rotate Red", rotredBlendFunc, "Amount to rotate", NULL, NULL },
		{ "Rotate Green", rotgreenBlendFunc, "Amount to rotate", NULL, NULL },
		{ "Rotate Blue", rotblueBlendFunc, "Amount to rotate", NULL, NULL },
		{ "Posterize", posterizeBlendFunc, "Amount to posertize", NULL, posterizeStartFunc },
		{ "Solarize", solarizeBlendFunc, "Threshold", NULL, NULL },
		{ NULL, NULL, NULL, NULL, NULL } // Must be here, represents the end of the list
	};